/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2011 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#include <Python.h>
#include <datetime.h>

#include <errno.h>
#include <time.h>
#include <unistd.h>

#include <pakfire/filelist.h>
#include <pakfire/package.h>
#include <pakfire/packagelist.h>

#include "file.h"
#include "package.h"
#include "util.h"

static unsigned long fibonnacci(const clock_t* deadline) {
	clock_t now = clock();

	unsigned long f1 = 1;
	unsigned long f2 = 1;

	// Count iterations
	unsigned long counter = 0;

	while (now < *deadline) {
		unsigned long next = f1 + f2;
		f1 = f2;
		f2 = next;

		now = clock();
		counter++;
	}

	return counter;
}

PyObject* performance_index(PyObject* self, PyObject* args) {
	int seconds = 1;

	if (!PyArg_ParseTuple(args, "|i", &seconds)) {
		return NULL;
	}

	if (seconds == 0) {
		PyErr_SetString(PyExc_ValueError, "Runtime must be one second or longer");
		return NULL;
	}

	// Determine the number of online processors
	int processors = sysconf(_SC_NPROCESSORS_ONLN);

	// Determine deadline
	clock_t deadline = clock();
	deadline += CLOCKS_PER_SEC * seconds;

	// Run Fibonnacci until deadline
	unsigned long iterations = fibonnacci(&deadline);

	// Times the result by the number of processors
	iterations *= processors;

	// Normalise to a second
	iterations /= seconds;

	return PyLong_FromUnsignedLong(iterations);
}

PyObject* PyList_FromPackageList(struct pakfire_packagelist* packagelist) {
	PyObject* list = PyList_New(0);

	int count = pakfire_packagelist_count(packagelist);
	for (int i = 0; i < count; i++) {
		struct pakfire_package* package = pakfire_packagelist_get(packagelist, i);

		PyObject* item = new_package(&PackageType, package);
		PyList_Append(list, item);

		pakfire_package_unref(package);
		Py_DECREF(item);
	}

	return list;
}

PyObject* PyList_FromFileList(struct pakfire_filelist* filelist) {
	struct pakfire_file* file = NULL;
	PyObject* f = NULL;
	int r;

	PyObject* list = PyList_New(0);
	if (!list)
		goto ERROR;

	// Fetch size
	const size_t size = pakfire_filelist_size(filelist);

	for (unsigned int i = 0; i < size; i++) {
		file = pakfire_filelist_get(filelist, i);
		if (!file)
			goto ERROR;

		// Create a new File object
		f = new_file(file);
		if (!f)
			goto ERROR;

		// Append the new object to the list
		r = PyList_Append(list, f);
		Py_DECREF(f);

		// If we could not append to the list, we will break
		if (r)
			goto ERROR;

		// Free file
		pakfire_file_unref(file);
	}

	return list;

ERROR:
	Py_XDECREF(list);
	return NULL;
}

PyObject* PyDateTime_FromTime_t(const time_t* t) {
	struct tm buffer;

	struct tm* tm = gmtime_r(t, &buffer);
	if (!tm)
		return NULL;

	PyDateTime_IMPORT;

	return PyDateTime_FromDateAndTime(
		1900 + tm->tm_year,
		1 + tm->tm_mon,
		tm->tm_mday,
		tm->tm_hour,
		tm->tm_min,
		tm->tm_sec,
		0
	);
}
