/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2014 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/sendfile.h>
#include <sys/stat.h>

// libarchive
#include <archive.h>
#include <archive_entry.h>

// JSON-C
#include <json.h>

#include <pakfire/archive.h>
#include <pakfire/compress.h>
#include <pakfire/dependencies.h>
#include <pakfire/digest.h>
#include <pakfire/file.h>
#include <pakfire/filelist.h>
#include <pakfire/i18n.h>
#include <pakfire/logging.h>
#include <pakfire/package.h>
#include <pakfire/pakfire.h>
#include <pakfire/parser.h>
#include <pakfire/private.h>
#include <pakfire/repo.h>
#include <pakfire/scriptlet.h>
#include <pakfire/string.h>
#include <pakfire/util.h>

#define MAX_SCRIPTLETS 9

struct pakfire_archive {
	struct pakfire* pakfire;
	int nrefs;

	char path[PATH_MAX];
	FILE* f;
	struct stat stat;

	struct pakfire_package* package;

	// metadata
	unsigned int format;
	struct json_object* metadata;
	struct pakfire_parser* parser;

	struct pakfire_filelist* filelist;

	// Scriptlets
	struct pakfire_scriptlet* scriptlets[MAX_SCRIPTLETS];
	unsigned int num_scriptlets;

	// Digests
	struct pakfire_digests digests;

	// Verify Status
	int verify;
};

static int pakfire_archive_compute_digests(struct pakfire_archive* archive) {
	int r;

	// Start reading at the beginning
	rewind(archive->f);

	// Calculate digest
	r = pakfire_digests_compute_from_file(archive->pakfire, &archive->digests,
			PAKFIRE_ARCHIVE_CHECKSUM, archive->f);
	if (r)
		ERROR(archive->pakfire, "Could not calculate digest of %s: %m\n", archive->path);

	return r;
}

/*
	A helper function to close the archive and reset our data structures
*/
static void close_archive(struct pakfire_archive* archive, struct archive* a) {
	if (a)
		archive_read_free(a);
}

/*
	A helper function that opens the archive for reading
*/
static struct archive* open_archive(struct pakfire_archive* archive) {
	// Create a new archive object
	struct archive* a = archive_read_new();
	if (!a)
		return NULL;

	// Archives must be uncompressed tarballs
	archive_read_support_format_tar(a);

	// Archives are either compressed using XZ or Zstandard
	archive_read_support_filter_xz(a);
	archive_read_support_filter_zstd(a);

	// Start reading from the beginning
	rewind(archive->f);

	// Try opening the archive file
	int r = archive_read_open_FILE(a, archive->f);
	if (r) {
		ERROR(archive->pakfire, "Could not open archive %s: %s\n",
			archive->path, archive_error_string(a));
		goto ERROR;
	}

	// Success
	return a;

ERROR:
	close_archive(archive, a);

	return NULL;
}

static int pakfire_archive_walk(struct pakfire_archive* archive,
		pakfire_walk_callback callback, pakfire_walk_filter_callback filter_callback, void* data) {
	int r;

	// Open the archive file
	struct archive* a = open_archive(archive);
	if (!a)
		return 1;

	// Walk through the archive
	r = pakfire_walk(archive->pakfire, a, callback, filter_callback, data);

	// Close the archive
	if (a)
		close_archive(archive, a);

	return r;
}

static int find_archive_entry(struct pakfire_archive* archive, struct archive_entry** entry,
		struct archive* a, const char* filename) {
	while (1) {
		int r = archive_read_next_header(a, entry);

		// Nothing found
		if (r == ARCHIVE_EOF) {
			break;

		// Some error has occured
		} else if (r)
			return r;

		// Compare the name
		const char* entry_name = archive_entry_pathname(*entry);

		// Match?
		if (strcmp(entry_name, filename) == 0)
			return 0;
	}

	DEBUG(archive->pakfire, "Could not find an entry named '%s'\n", filename);
	errno = ENOENT;

	// Nothing found
	*entry = NULL;
	return 1;
}

static la_ssize_t pakfire_archive_read_callback(struct archive* a,
		void* client_data, const void** buffer) {
	struct archive* archive = (struct archive*)client_data;

	size_t len = 0;
	off_t offset = 0;

	// Try reading the next block (without copying it)
	int r = archive_read_data_block(archive, buffer, &len, &offset);

	switch (r) {
		case ARCHIVE_OK:
			return len;

		// Return zero to signal that everything was read
		case ARCHIVE_EOF:
			return 0;

		// Return -1 on any other errors
		default:
			return -1;
	}
}

static struct archive* pakfire_archive_open_payload(struct pakfire_archive* archive,
		struct archive* a, size_t* size) {
	struct archive_entry* entry = NULL;
	struct archive* payload = NULL;

	// Find the payload
	int r = find_archive_entry(archive, &entry, a, "data.img");
	if (r)
		goto ERROR;

	// Store size
	if (size)
		*size = archive_entry_size(entry);

	// Allocate a new archive object
	payload = archive_read_new();
	if (!payload)
		goto ERROR;

	// All of our packages are tar balls
	archive_read_support_format_tar(payload);

	// They are compressed using XZ
	archive_read_support_filter_xz(payload);

	// Try opening the payload archive
	r = archive_read_open2(payload, a, NULL, pakfire_archive_read_callback, NULL, NULL);
	if (r) {
		ERROR(archive->pakfire, "Could not open payload archive: %s\n",
			archive_error_string(payload));
		goto ERROR;
	}

	return payload;

ERROR:
	if (payload)
		archive_read_free(payload);

	return NULL;
}

static void pakfire_archive_free(struct pakfire_archive* archive) {
	// Close the file
	if (archive->f)
		fclose(archive->f);

	// Free scriptlets
	for (unsigned int i = 0; i < archive->num_scriptlets; i++)
		pakfire_scriptlet_unref(archive->scriptlets[i]);

	if (archive->filelist)
		pakfire_filelist_unref(archive->filelist);
	if (archive->package)
		pakfire_package_unref(archive->package);
	if (archive->metadata)
		json_object_put(archive->metadata);
	if (archive->parser)
		pakfire_parser_unref(archive->parser);
	pakfire_unref(archive->pakfire);
	free(archive);
}

static int pakfire_archive_create(struct pakfire_archive** archive, struct pakfire* pakfire) {
	struct pakfire_archive* a = calloc(1, sizeof(*a));
	if (!a)
		return ENOMEM;

	a->pakfire = pakfire_ref(pakfire);
	a->nrefs = 1;

	*archive = a;
	return 0;
}

PAKFIRE_EXPORT struct pakfire_archive* pakfire_archive_ref(struct pakfire_archive* archive) {
	++archive->nrefs;

	return archive;
}

PAKFIRE_EXPORT struct pakfire_archive* pakfire_archive_unref(struct pakfire_archive* archive) {
	if (--archive->nrefs > 0)
		return archive;

	pakfire_archive_free(archive);

	return NULL;
}

static struct pakfire_package* pakfire_archive_get_package(struct pakfire_archive* archive) {
	if (!archive->package) {
		int r = pakfire_archive_make_package(archive, NULL, &archive->package);
		if (r)
			return NULL;
	}

	return pakfire_package_ref(archive->package);
}

// Metadata

static int pakfire_archive_parse_json_metadata(struct pakfire_archive* archive,
		const char* data, const size_t length) {
	int r = 1;

	// Create tokener
	struct json_tokener* tokener = json_tokener_new();
	if (!tokener) {
		ERROR(archive->pakfire, "Could not allocate JSON tokener: %m\n");
		goto ERROR;
	}

	// Parse JSON from buffer
	archive->metadata = json_tokener_parse_ex(tokener, data, length);
	if (!archive->metadata) {
		enum json_tokener_error error = json_tokener_get_error(tokener);

		ERROR(archive->pakfire, "JSON parsing error: %s\n",
			json_tokener_error_desc(error));
		goto ERROR;
	}

	DEBUG(archive->pakfire, "Successfully parsed package metadata:\n%s\n",
		json_object_to_json_string_ext(archive->metadata,
			JSON_C_TO_STRING_PRETTY|JSON_C_TO_STRING_PRETTY_TAB));

	// Success
	r = 0;

ERROR:
	if (tokener)
		json_tokener_free(tokener);

	return r;
}

static int pakfire_archive_parse_legacy_metadata(struct pakfire_archive* archive,
		const char* data, const size_t length) {
	int r = 1;

	// Allocate a new parser
	archive->parser = pakfire_parser_create(archive->pakfire, NULL, NULL, 0);
	if (!archive->parser)
		goto ERROR;

	// Parse the metadata
	r = pakfire_parser_parse_data(archive->parser, data, length, NULL);
	if (r) {
		ERROR(archive->pakfire, "Error parsing metadata\n");
		goto ERROR;
	}

	// Success
	return 0;

ERROR:
	if (archive->parser) {
		pakfire_parser_unref(archive->parser);
		archive->parser = NULL;
	}

	return r;
}

static int pakfire_archive_parse_format(struct pakfire_archive* archive,
		const char* data, const size_t length) {
	// Check if format has already been set
	if (archive->format) {
		ERROR(archive->pakfire, "Archive format has already been parsed\n");
		errno = EINVAL;
		return 1;
	}

	// Parse the format
	archive->format = strtoul(data, NULL, 10);

	switch (archive->format) {
		// Handle all supported formats
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
			break;

		// Break on anything else
		default:
			ERROR(archive->pakfire, "This version of Pakfire does not support "
				"archive format %d\n", archive->format);
			errno = EINVAL;
			return 1;
	}

	DEBUG(archive->pakfire, "Archive format is %d\n", archive->format);

	return 0;
}

static int pakfire_archive_parse_scriptlet(struct pakfire_archive* archive,
		const char* path, const char* data, const size_t length) {
	struct pakfire_scriptlet* scriptlet = NULL;
	const char* type = NULL;
	int r;

	// Check for any available space
	if (archive->num_scriptlets >= MAX_SCRIPTLETS) {
		ERROR(archive->pakfire, "Too many scriptlets\n");
		errno = ENOBUFS;
		return 1;
	}

	// Determine type
	if (archive->format >= 6)
		type = pakfire_path_relpath(".scriptlets/", path);
	else if (archive->format >= 1)
		type = pakfire_path_relpath("scriptlet/", path);

	if (!type) {
		ERROR(archive->pakfire, "Could not determine the scriptlet type from '%s'\n", path);
		errno = EINVAL;
		return 1;
	}

	// Allocate a scriptlet
	r = pakfire_scriptlet_create(&scriptlet, archive->pakfire, type, data, length);
	if (r)
		return r;

	// Store scriptlet
	archive->scriptlets[archive->num_scriptlets++] = scriptlet;

	return 0;
}

static int __pakfire_archive_read_metadata(struct pakfire* pakfire, struct archive* a,
		struct archive_entry* entry, void* p) {
	struct pakfire_archive* archive = (struct pakfire_archive*)p;

	char* data = NULL;
	size_t length = 0;
	int r;

	const char* path = archive_entry_pathname(entry);

	DEBUG(pakfire, "Reading metadata file: %s\n", path);

	// Load the file into memory
	r = pakfire_archive_copy_data_to_buffer(archive->pakfire, a, entry, &data, &length);
	if (r) {
		ERROR(archive->pakfire, "Could not read data from archive: %s\n",
			archive_error_string(a));
		goto ERROR;
	}

	// Format >= 6
	if (archive->format >= 6) {
		// Parse PKGINFO
		if (strcmp(path, ".PKGINFO") == 0) {
			r = pakfire_archive_parse_json_metadata(archive, data, length);
			if (r)
				goto ERROR;

		// Parse scriptlets
		} else if (pakfire_string_startswith(path, ".scriptlets/")) {
			r = pakfire_archive_parse_scriptlet(archive, path, data, length);
			if (r)
				goto ERROR;
		}

	// Format >= 1
	} else if (archive->format >= 1) {
		// Parse info
		if (strcmp(path, "info") == 0) {
			r = pakfire_archive_parse_legacy_metadata(archive, data, length);
			if (r)
				goto ERROR;

		// Parse scriptlets
		} else if (pakfire_string_startswith(path, "scriptlet/")) {
			r = pakfire_archive_parse_scriptlet(archive, path, data, length);
			if (r)
				goto ERROR;
		}

	// pakfire-format
	} else if (strcmp(path, "pakfire-format") == 0) {
		r = pakfire_archive_parse_format(archive, data, length);
		if (r)
			goto ERROR;
	}

ERROR:
	if (data)
		free(data);

	return r;
}

static int __pakfire_archive_filter_metadata(struct pakfire* pakfire,
		struct archive* a, struct archive_entry* entry, void* p) {
	struct pakfire_archive* archive = (struct pakfire_archive*)p;

	const char* path = archive_entry_pathname(entry);

	// Format >= 6
	if (archive->format >= 6) {
		// Anything that starts with "." is a metadata file
		if (*path == '.')
			return PAKFIRE_WALK_OK;

		// Otherwise, the payload begins
		return PAKFIRE_WALK_DONE;

	// Format >= 1
	} else if (archive->format >= 1) {
		// info
		if (strcmp(path, "info") == 0)
			return PAKFIRE_WALK_OK;

		// scriptlets
		else if (pakfire_string_startswith(path, "scriptlets/"))
			return PAKFIRE_WALK_OK;

		// Ignore anything else
		return PAKFIRE_WALK_SKIP;

	// The pakfire-format file is part of the metadata
	} else if (strcmp(path, "pakfire-format") == 0) {
		return PAKFIRE_WALK_OK;
	}

	// Unknown file
	return PAKFIRE_WALK_ERROR;
}

static int pakfire_archive_read_metadata(struct pakfire_archive* archive) {
	int r;

	// Check if the archive file actually has any contect
	if (!archive->stat.st_size) {
		ERROR(archive->pakfire, "Trying to open an empty archive file\n");
		errno = EINVAL;
		return 1;
	}

	DEBUG(archive->pakfire, "Reading archive metadata...\n");

	// Walk through the archive
	r = pakfire_archive_walk(archive, __pakfire_archive_read_metadata,
		__pakfire_archive_filter_metadata, archive);
	if (r)
		return r;

	// Check if we could successfully read something
	if (!archive->format) {
		ERROR(archive->pakfire, "Archive has an unknown format\n");
		errno = EINVAL;
		return 1;
	}

	// Check if we have read some metadata
	if (!archive->metadata && !archive->parser) {
		ERROR(archive->pakfire, "Archive has no metadata\n");
		errno = EINVAL;
		return 1;
	}

	return 0;
}

static int pakfire_archive_try_open(struct pakfire_archive* archive, const char* path) {
	int r;

	if (!path) {
		errno = EINVAL;
		return 1;
	}

	DEBUG(archive->pakfire, "Opening archive %s\n", path);

	// Store path
	pakfire_string_set(archive->path, path);

	// Open the file (and keep the file descriptor open)
	archive->f = fopen(archive->path, "r");
	if (!archive->f)
		return 1;

	// Let the kernel know, that we will read the file sequentially
	r = posix_fadvise(fileno(archive->f), 0, 0, POSIX_FADV_SEQUENTIAL);
	if (r) {
		ERROR(archive->pakfire, "posix_fadvise() failed: %m\n");
		goto ERROR;
	}

	// Call stat() on f
	r = fstat(fileno(archive->f), &archive->stat);
	if (r) {
		ERROR(archive->pakfire, "Could not stat archive: %m\n");
		goto ERROR;
	}

	// Read all package metadata
	r = pakfire_archive_read_metadata(archive);
	if (r) {
		ERROR(archive->pakfire, "Could not open archive: %m\n");
		goto ERROR;
	}

ERROR:
	return r;
}

PAKFIRE_EXPORT int pakfire_archive_open(struct pakfire_archive** archive, struct pakfire* pakfire, const char* path) {
	int r = pakfire_archive_create(archive, pakfire);
	if (r)
		return r;

	r = pakfire_archive_try_open(*archive, path);
	if (r)
		goto ERROR;

	return 0;

ERROR:
	pakfire_archive_unref(*archive);
	*archive = NULL;

	return r;
}

static struct json_object* pakfire_archive_metadata_get_object(
		struct pakfire_archive* archive, const char* key1, const char* key2) {
	struct json_object* object = archive->metadata;
	int r;

	const char* keys[] = {
		key1,
		key2,
		NULL,
	};

	// Walk through all keys
	for (const char** key = keys; *key; key++) {
		// Try finding a matching JSON object
		r = json_object_object_get_ex(object, *key, &object);
		if (!r) {
			DEBUG(archive->pakfire, "Could not find JSON object at '%s': %m\n", *key);
			break;
		}
	}

	return object;
}

static const char* pakfire_archive_metadata_get(
		struct pakfire_archive* archive, const char* key1, const char* key2) {
	// Try finding an object
	struct json_object* object = pakfire_archive_metadata_get_object(archive, key1, key2);
	if (!object)
		return NULL;

	// Return the object as string
	return json_object_get_string(object);
}

static int64_t pakfire_archive_metadata_get_int64(
		struct pakfire_archive* archive, const char* key1, const char* key2) {
	// Try finding an object
	struct json_object* object = pakfire_archive_metadata_get_object(archive, key1, key2);
	if (!object)
		return 0;

	// Return the object as integer
	return json_object_get_int64(object);
}

PAKFIRE_EXPORT char* pakfire_archive_get(struct pakfire_archive* archive,
		const char* namespace, const char* key) {
	return pakfire_parser_get(archive->parser, namespace, key);
}

static int __pakfire_archive_read_legacy(struct pakfire_archive* archive,
		const char* path, char** data, size_t* length) {
	int r = 1;

	struct archive* a = NULL;
	struct archive_entry* entry = NULL;
	struct archive* payload = NULL;

	// Open the archive
	a = open_archive(archive);
	if (!a) {
		r = 1;
		goto ERROR;
	}

	payload = pakfire_archive_open_payload(archive, a, NULL);
	if (!payload)
		goto ERROR;

	r = find_archive_entry(archive, &entry, payload, path);
	if (r)
		goto ERROR;

	r = pakfire_archive_copy_data_to_buffer(archive->pakfire, payload, entry,
		data, length);

ERROR:
	if (payload)
		archive_read_free(payload);
	close_archive(archive, a);

	return r;
}

struct pakfire_archive_read {
	struct pakfire_archive* archive;
	const char* path;
	char** data;
	size_t* length;
};

static int __pakfire_archive_filter_payload(struct pakfire* pakfire,
		struct archive* a, struct archive_entry* entry, void* p) {
	struct pakfire_archive_read* _read = (struct pakfire_archive_read*)p;
	struct pakfire_archive* archive = _read->archive;

	const char* path = NULL;

	// Format >= 6
	if (archive->format >= 6) {
		path = archive_entry_pathname(entry);

		// Skip any hidden files
		if (path && *path == '.')
			return ARCHIVE_RETRY;
	}

	// Otherwise permit reading any files
	return ARCHIVE_OK;
}

static int __pakfire_archive_read_payload(struct pakfire* pakfire,
		struct archive* a, struct archive_entry* entry, void* p) {
	struct pakfire_archive_read* _read = (struct pakfire_archive_read*)p;
	int r;

	const char* path = archive_entry_pathname(entry);

	if (strcmp(_read->path, path) == 0) {
		r = pakfire_archive_copy_data_to_buffer(pakfire, a, entry,
			_read->data, _read->length);
		if (r)
			return r;
	}

	return 0;
}

static int __pakfire_archive_read(struct pakfire_archive* archive,
		const char* path, char** data, size_t* length) {
	struct pakfire_archive_read _read = {
		.path   = path,
		.data   = data,
		.length = length,
	};

	// Walk through the archive
	return pakfire_archive_walk(archive, __pakfire_archive_read_payload,
		__pakfire_archive_filter_payload, &_read);
}

PAKFIRE_EXPORT int pakfire_archive_read(struct pakfire_archive* archive,
		const char* path, char** data, size_t* length) {
	// Strip leading / from filenames, because tarballs don't use any leading slashes
	path = pakfire_path_relpath("/", path);

	// Check inputs
	if (!path || !data || !length) {
		errno = EINVAL;
		return 1;
	}

	if (archive->format >= 6)
		return __pakfire_archive_read(archive, path, data, length);
	else
		return __pakfire_archive_read_legacy(archive, path, data, length);
}

int pakfire_archive_copy(struct pakfire_archive* archive, const char* path) {
	if (!path) {
		errno = EINVAL;
		return 1;
	}

	// Determine the file size
	ssize_t size = pakfire_archive_get_size(archive);
	if (size < 0)
		return 1;

	DEBUG(archive->pakfire, "Copying %s to %s...\n", archive->path, path);

	// Ensure we copy from the very beginning
	rewind(archive->f);

	// Open destination file
	FILE* f = fopen(path, "w");
	if (!f)
		return 1;

	int r = 1;

	// Copy everything
	ssize_t bytes_written = sendfile(fileno(f), fileno(archive->f), NULL, size);
	if (bytes_written < size) {
		ERROR(archive->pakfire, "Could not copy archive (%zu byte(s) written): %m\n",
			bytes_written);
		goto ERROR;
	}

	// Success
	r = 0;

ERROR:
	fclose(f);

	// Delete the file on error
	if (r)
		unlink(path);

	return r;
}

static int __pakfire_archive_extract(struct pakfire_archive* archive, int flags) {
	struct pakfire_filelist* filelist = NULL;
	struct pakfire_package* pkg = NULL;
	struct archive* a = NULL;
	struct archive* payload = NULL;
	char prefix[PATH_MAX] = "";
	size_t size = 0;
	int r = 1;

	// Fetch package
	pkg = pakfire_archive_get_package(archive);
	if (!pkg)
		goto ERROR;

	// Fetch NEVRA
	const char* nevra = pakfire_package_get_string(pkg, PAKFIRE_PKG_NEVRA);

	DEBUG(archive->pakfire, "Extracting %s\n", archive->path);

	// Set prefix for source packages
	if (pakfire_package_is_source(pkg)) {
		r = pakfire_string_format(prefix, "/usr/src/packages/%s", nevra);
		if (r)
			goto ERROR;
	}

	// Load the filelist (if not done already)
	if (!archive->filelist) {
		r = pakfire_filelist_create(&filelist, archive->pakfire);
		if (r)
			goto ERROR;
	}

	// Open the archive
	a = open_archive(archive);
	if (!a) {
		r = 1;
		goto ERROR;
	}

	// Open the internal tarball with the payload for older formats
	if (archive->format < 6) {
		payload = pakfire_archive_open_payload(archive, a, &size);
		if (!payload)
			goto ERROR;
	}

	// Extract
	r = pakfire_extract(archive->pakfire, (payload) ? payload : a, size,
		filelist, prefix, nevra, flags);
	if (r)
		goto ERROR;

	// Store the filelist permanently
	if (!archive->filelist)
		archive->filelist = pakfire_filelist_ref(filelist);

ERROR:
	if (filelist)
		pakfire_filelist_unref(filelist);
	if (pkg)
		pakfire_package_unref(pkg);
	if (payload)
		archive_read_free(payload);
	if (a)
		close_archive(archive, a);

	return r;
}

PAKFIRE_EXPORT int pakfire_archive_extract(struct pakfire_archive* archive) {
	return __pakfire_archive_extract(archive, 0);
}

PAKFIRE_EXPORT const char* pakfire_archive_get_path(struct pakfire_archive* archive) {
	return archive->path;
}

PAKFIRE_EXPORT unsigned int pakfire_archive_get_format(struct pakfire_archive* archive) {
	return archive->format;
}

static int pakfire_archive_load_filelist(struct pakfire_archive* archive) {
	// Perform a dry-run extraction
	return __pakfire_archive_extract(archive,
		PAKFIRE_EXTRACT_DRY_RUN|PAKFIRE_EXTRACT_NO_PROGRESS);
}

PAKFIRE_EXPORT struct pakfire_filelist* pakfire_archive_get_filelist(struct pakfire_archive* archive) {
	if (!archive->filelist) {
		int r = pakfire_archive_load_filelist(archive);
		if (r)
			return NULL;
	}

	if (!archive->filelist)
		return NULL;

	return pakfire_filelist_ref(archive->filelist);
}

PAKFIRE_EXPORT int pakfire_archive_verify(struct pakfire_archive* archive, int* status) {
	// XXX currently not implemented
	return 0;
}

PAKFIRE_EXPORT ssize_t pakfire_archive_get_size(struct pakfire_archive* archive) {
	return archive->stat.st_size;
}

int pakfire_archive_check_digest(struct pakfire_archive* archive,
		const enum pakfire_digest_types type, const unsigned char* digest, const size_t length) {
	size_t computed_length = 0;
	int r;

	// Compute the digest
	r = pakfire_archive_compute_digests(archive);
	if (r)
		return r;

	// Compare computed with expected digest
	r = pakfire_digests_compare_one(archive->pakfire, &archive->digests,
		type, digest, length);
	if (r) {
		const unsigned char* computed_digest = pakfire_digest_get(
			&archive->digests, type, &computed_length);

		char* expected_hexdigest = __pakfire_hexlify(digest, length);
		char* computed_hexdigest = __pakfire_hexlify(computed_digest, computed_length);

		ERROR(archive->pakfire, "Archive digest does not match for %s:\n", archive->path);
		ERROR(archive->pakfire, "  Expected: %s\n", expected_hexdigest);
		ERROR(archive->pakfire, "  Computed: %s\n", computed_hexdigest);

		if (expected_hexdigest)
			free(expected_hexdigest);
		if (computed_hexdigest)
			free(computed_hexdigest);

		return r;
	}

	return r;
}

static int pakfire_archive_make_filelist_from_json(struct pakfire_archive* archive,
		struct pakfire_filelist** filelist) {
	struct json_object* array = NULL;
	int r;

	struct pakfire_filelist* list = NULL;
	struct pakfire_file* file = NULL;

	// Fetch the array with the filelist
	array = pakfire_archive_metadata_get_object(archive, "filelist", NULL);
	if (!array) {
		ERROR(archive->pakfire, "Archive has no filelist: %m\n");
		return 1;
	}

	// Determine the length of the array
	const size_t length = json_object_array_length(array);

	// End here if the array is empty
	if (!length)
		return 0;

	// Create a new filelist object
	r = pakfire_filelist_create(&list, archive->pakfire);
	if (r)
		goto ERROR;

	// Walk through all items in this array
	for (unsigned int i = 0; i < length; i++) {
		struct json_object* item = json_object_array_get_idx(array, i);
		if (!item)
			continue;

		// Extract the path value
		const char* path = json_object_get_string(item);
		if (!path)
			continue;

		// Create a new file object
		r = pakfire_file_create(&file, archive->pakfire);
		if (r)
			goto ERROR;

		// Set path
		r = pakfire_file_set_path(file, path);
		if (r) {
			pakfire_file_unref(file);
			goto ERROR;
		}

		// Append the file to the filelist
		r = pakfire_filelist_append(list, file);
		if (r) {
			pakfire_file_unref(file);
			goto ERROR;
		}

		// Cleanup
		pakfire_file_unref(file);
	}

	// Reference the filelist
	*filelist = pakfire_filelist_ref(list);

ERROR:
	if (list)
		pakfire_filelist_unref(list);

	return r;
}

static int pakfire_archive_make_package_from_json(struct pakfire_archive* archive,
		struct pakfire_repo* repo, struct pakfire_package** package) {
	struct pakfire_package* pkg = NULL;
	struct pakfire_filelist* filelist = NULL;
	int r;

	// Calculate digest
	r = pakfire_archive_compute_digests(archive);
	if (r)
		return r;

	// Fetch the most basic package information
	const char* name = pakfire_archive_metadata_get(archive, "name", NULL);
	const char* evr  = pakfire_archive_metadata_get(archive, "evr", NULL);
	const char* arch = pakfire_archive_metadata_get(archive, "arch", NULL);

	// Create a new package object
	r = pakfire_package_create(&pkg, archive->pakfire, repo, name, evr, arch);
	if (r)
		return r;

#ifdef ENABLE_DEBUG
	const char* nevra = pakfire_package_get_string(pkg, PAKFIRE_PKG_NEVRA);

	DEBUG(archive->pakfire, "Created package %s (%p) from archive %p\n",
		nevra, pkg, archive);
#endif

	// Set path
	pakfire_package_set_string(pkg, PAKFIRE_PKG_PATH, archive->path);

	// Set digest
	switch (PAKFIRE_ARCHIVE_CHECKSUM) {
		case PAKFIRE_DIGEST_SHA2_512:
			pakfire_package_set_digest(pkg, PAKFIRE_ARCHIVE_CHECKSUM,
				archive->digests.sha2_512, sizeof(archive->digests.sha2_512));
			break;

		case PAKFIRE_DIGEST_SHA2_256:
			pakfire_package_set_digest(pkg, PAKFIRE_ARCHIVE_CHECKSUM,
				archive->digests.sha2_256, sizeof(archive->digests.sha2_256));
			break;

		case PAKFIRE_DIGEST_UNDEFINED:
			r = 1;
			goto ERROR;
	}

	// Vendor
	const char* vendor = pakfire_archive_metadata_get(archive, "vendor", NULL);
	if (vendor) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_VENDOR, vendor);
		if (r)
			goto ERROR;
	}

	// UUID
	const char* uuid = pakfire_archive_metadata_get(archive, "uuid", NULL);
	if (uuid) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_UUID, uuid);
		if (r)
			goto ERROR;
	}

	// Groups
	const char* groups = pakfire_archive_metadata_get(archive, "groups", NULL);
	if (groups) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_GROUPS, groups);
		if (r)
			goto ERROR;
	}

	// Distribution
	const char* distro = pakfire_archive_metadata_get(archive, "distribution", NULL);
	if (distro) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_DISTRO, distro);
		if (r)
			goto ERROR;
	}

	// Packager
	const char* packager = pakfire_archive_metadata_get(archive, "packager", NULL);
	if (packager) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_PACKAGER, packager);
		if (r)
			goto ERROR;
	}

	// URL
	const char* url = pakfire_archive_metadata_get(archive, "url", NULL);
	if (url) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_URL, url);
		if (r)
			goto ERROR;
	}

	// License
	const char* license = pakfire_archive_metadata_get(archive, "license", NULL);
	if (license) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_LICENSE, license);
		if (r)
			goto ERROR;
	}

	// Summary
	const char* summary = pakfire_archive_metadata_get(archive, "summary", NULL);
	if (summary) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_SUMMARY, summary);
		if (r)
			goto ERROR;
	}

	// Description
	const char* description = pakfire_archive_metadata_get(archive, "description", NULL);
	if (description) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_DESCRIPTION, description);
		if (r)
			goto ERROR;
	}

	// Installed package size
	size_t installsize = pakfire_archive_metadata_get_int64(archive, "size", NULL);
	if (installsize) {
		r = pakfire_package_set_num(pkg, PAKFIRE_PKG_INSTALLSIZE, installsize);
		if (r)
			goto ERROR;
	}

	// Build Host
	const char* build_host = pakfire_archive_metadata_get(archive, "build", "host");
	if (build_host) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_BUILD_HOST, build_host);
		if (r)
			goto ERROR;
	}

	// Build ID
	const char* build_id = pakfire_archive_metadata_get(archive, "build", "id");
	if (build_id) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_BUILD_ID, build_id);
		if (r)
			goto ERROR;
	}

	// Build Time
	time_t build_time = pakfire_archive_metadata_get_int64(archive, "build", "time");
	if (build_time) {
		r = pakfire_package_set_num(pkg, PAKFIRE_PKG_BUILD_TIME, build_time);
		if (r)
			goto ERROR;
	}

	// Source package
	const char* source_name = pakfire_archive_metadata_get(archive, "build", "source-name");
	if (source_name) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_SOURCE_NAME, source_name);
		if (r)
			goto ERROR;
	}

	// Source EVR
	const char* source_evr = pakfire_archive_metadata_get(archive, "build", "source-evr");
	if (source_evr) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_SOURCE_EVR, source_evr);
		if (r)
			goto ERROR;
	}

	// Source arch
	const char* source_arch = pakfire_archive_metadata_get(archive, "build", "source-arch");
	if (source_arch) {
		r = pakfire_package_set_string(pkg, PAKFIRE_PKG_SOURCE_ARCH, source_arch);
		if (r)
			goto ERROR;
	}

	// Dependencies
	for (const struct pakfire_dep* dep = pakfire_deps; dep->key; dep++) {
		struct json_object* array = pakfire_archive_metadata_get_object(
			archive, "dependencies", dep->name);
		if (!array)
			continue;

		// Determine the length of the array
		const size_t length = json_object_array_length(array);
		if (!length)
			continue;

		// Walk through all items in this array
		for (unsigned int i = 0; i < length; i++) {
			struct json_object* item = json_object_array_get_idx(array, i);
			if (!item)
				continue;

			// Extract the string value
			const char* string = json_object_get_string(item);
			if (!string)
				continue;

			// Add the dependency to the package
			r = pakfire_package_add_dep(pkg, dep->key, string);
			if (r)
				goto ERROR;
		}
	}

	// Fetch the filelist
	r = pakfire_archive_make_filelist_from_json(archive, &filelist);
	if (r)
		goto ERROR;

	// Store the filelist
	if (filelist) {
		r = pakfire_package_set_filelist(pkg, filelist);
		if (r)
			goto ERROR;
	}

	// Success!
	*package = pkg;
	r = 0;

ERROR:
	if (filelist)
		pakfire_filelist_unref(filelist);

	return r;
}

static int pakfire_archive_package_set_string(struct pakfire_archive* archive,
		struct pakfire_package* pkg, const enum pakfire_package_key key,
		const char* key1, const char* key2) {
	char* value = NULL;
	int r;

	// Fetch the value
	value = pakfire_archive_get(archive, key1, key2);

	// Set the value
	r = pakfire_package_set_string(pkg, key, value);

	if (value)
		free(value);

	return r;
}

static int pakfire_archive_make_legacy_package(struct pakfire_archive* archive,
		struct pakfire_repo* repo, struct pakfire_package** package) {
	int r;

	// Calculate digest
	r = pakfire_archive_compute_digests(archive);
	if (r)
		return r;

	char* name = pakfire_archive_get(archive, "package", "name");
	char* arch = pakfire_archive_get(archive, "package", "arch");

	char* evr = pakfire_archive_get(archive, "package", "evr");
	if (!evr) {
		char* epoch   = pakfire_archive_get(archive, "package", "epoch");
		char* version = pakfire_archive_get(archive, "package", "version");
		char* release = pakfire_archive_get(archive, "package", "release");

		evr = pakfire_package_join_evr(epoch, version, release);

		if (epoch)
			free(epoch);
		if (version)
			free(version);
		if (release)
			free(release);
	}

	/*
		Check architecture
		Legacy package formats use "arch = all" for source packages.
		We need to check for "type = source" and then reset arch
	*/
	if (arch && strcmp(arch, "all") == 0) {
		char* type = pakfire_archive_get(archive, "package", "type");
		if (type) {
			if (strcmp(type, "source") == 0) {
				free(arch);
				arch = NULL;
			}
		}
		free(type);
	}

	struct pakfire_package* pkg = NULL;

	r = pakfire_package_create(&pkg, archive->pakfire, repo, name, evr,
		(arch) ? arch : "src");

	if (name)
		free(name);
	if (evr)
		free(evr);
	if (arch)
		free(arch);

	if (r)
		goto ERROR;

	// Set package pointer right here to avoid an endless loop when loading the filelist
	*package = pkg;

#ifdef ENABLE_DEBUG
	const char* nevra = pakfire_package_get_string(pkg, PAKFIRE_PKG_NEVRA);

	DEBUG(archive->pakfire, "Created package %s (%p) from archive %p\n",
		nevra, pkg, archive);
#endif

	// Set path
	pakfire_package_set_string(pkg, PAKFIRE_PKG_PATH, archive->path);

	// Set UUID
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_UUID, "package", "uuid");
	if (r)
		goto ERROR;

	// Set groups
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_GROUPS, "package", "groups");
	if (r)
		goto ERROR;

	// Set packager
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_PACKAGER, "package", "packager");
	if (r)
		goto ERROR;

	// Set URL
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_URL, "package", "url");
	if (r)
		goto ERROR;

	// Set license
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_LICENSE, "package", "license");
	if (r)
		goto ERROR;

	// Set summary
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_SUMMARY, "package", "summary");
	if (r)
		goto ERROR;

	// Set description
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_DESCRIPTION, "package", "description");
	if (r)
		goto ERROR;

	// Get package size
	pakfire_package_set_num(pkg, PAKFIRE_PKG_DOWNLOADSIZE,
		pakfire_archive_get_size(archive));

	// Get install size
	char* size = pakfire_archive_get(archive, "package", "size");
	if (size) {
		size_t s = strtoul(size, NULL, 10);
		free(size);

		pakfire_package_set_num(pkg, PAKFIRE_PKG_INSTALLSIZE, s);
	}

	// Set vendor
	r = pakfire_archive_package_set_string(archive, pkg,
			PAKFIRE_PKG_VENDOR, "distribution", "vendor");
	if (r)
		goto ERROR;

	// Set build host
	char* build_host = pakfire_archive_get(archive, "build", "host");
	if (build_host) {
		pakfire_package_set_string(pkg, PAKFIRE_PKG_BUILD_HOST, build_host);
		free(build_host);
	}

	// Set build ID
	char* build_id = pakfire_archive_get(archive, "build", "id");
	if (build_id) {
		pakfire_package_set_string(pkg, PAKFIRE_PKG_BUILD_ID, build_id);
		free(build_id);
	}

	// Set build time
	char* build_time = pakfire_archive_get(archive, "build", "time");
	if (build_time) {
		time_t t = strtoull(build_time, NULL, 10);
		free(build_time);

		pakfire_package_set_num(pkg, PAKFIRE_PKG_BUILD_TIME, t);
	}

	// Dependencies
	for (const struct pakfire_dep* dep = pakfire_deps; dep->key; dep++) {
		char* rels = pakfire_archive_get(archive, "dependencies", dep->name);
		if (rels) {
			r = pakfire_str2deps(archive->pakfire, pkg, dep->key, rels);
			free(rels);

			if (r)
				goto ERROR;
		}
	}

	// Import filelist
	struct pakfire_filelist* filelist = pakfire_archive_get_filelist(archive);
	if (filelist) {
		pakfire_package_set_filelist(pkg, filelist);
		pakfire_filelist_unref(filelist);
	}

	// Set digest
	switch (PAKFIRE_ARCHIVE_CHECKSUM) {
		case PAKFIRE_DIGEST_SHA2_512:
			pakfire_package_set_digest(pkg, PAKFIRE_ARCHIVE_CHECKSUM,
				archive->digests.sha2_512, sizeof(archive->digests.sha2_512));
			break;

		case PAKFIRE_DIGEST_SHA2_256:
			pakfire_package_set_digest(pkg, PAKFIRE_ARCHIVE_CHECKSUM,
				archive->digests.sha2_256, sizeof(archive->digests.sha2_256));
			break;

		case PAKFIRE_DIGEST_UNDEFINED:
			break;
	}

ERROR:
	return r;
}

/*
	Copy all metadata from this archive to the package object
*/
PAKFIRE_EXPORT int pakfire_archive_make_package(struct pakfire_archive* archive,
		struct pakfire_repo* repo, struct pakfire_package** package) {
	struct pakfire_repo* dummy = NULL;
	int r;

	// Use dummy repo if no repository was passed
	if (!repo) {
		dummy = pakfire_get_repo(archive->pakfire, PAKFIRE_REPO_DUMMY);
		if (!dummy)
			return 1;

		repo = dummy;
	}

	// Make package from JSON metadata
	if (archive->format >= 6)
		r = pakfire_archive_make_package_from_json(archive, repo, package);

	// Use legacy stuff
	else
		r = pakfire_archive_make_legacy_package(archive, repo, package);

	// Free dummy repository
	if (dummy)
		pakfire_repo_unref(dummy);

	return r;
}

struct pakfire_scriptlet* pakfire_archive_get_scriptlet(
		struct pakfire_archive* archive, const char* type) {
	struct pakfire_scriptlet* scriptlet = NULL;

	for (unsigned int i = 0; i < archive->num_scriptlets; i++) {
		scriptlet = archive->scriptlets[i];

		// Fetch type
		const char* t = pakfire_scriptlet_get_type(scriptlet);

		// Compare type
		if (strcmp(t, type) == 0)
			return pakfire_scriptlet_ref(scriptlet);
	}

	return NULL;
}
