/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2014 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#include <errno.h>
#include <fnmatch.h>
#include <fts.h>
#include <stdlib.h>
#include <string.h>

#include <archive.h>

#include <pakfire/file.h>
#include <pakfire/filelist.h>
#include <pakfire/i18n.h>
#include <pakfire/logging.h>
#include <pakfire/pakfire.h>
#include <pakfire/private.h>
#include <pakfire/progressbar.h>
#include <pakfire/string.h>
#include <pakfire/util.h>

struct pakfire_filelist {
	struct pakfire* pakfire;
	int nrefs;

	struct pakfire_file** elements;
	size_t elements_size;

	size_t size;
};

static int pakfire_filelist_grow(struct pakfire_filelist* list, size_t size) {
	struct pakfire_file** elements = reallocarray(list->elements,
		list->elements_size + size, sizeof(*list->elements));
	if (!elements)
		return -errno;

	list->elements = elements;
	list->elements_size += size;

	return 0;
}

PAKFIRE_EXPORT int pakfire_filelist_create(struct pakfire_filelist** list, struct pakfire* pakfire) {
	struct pakfire_filelist* l = calloc(1, sizeof(*l));
	if (!l)
		return -ENOMEM;

	l->pakfire = pakfire_ref(pakfire);
	l->nrefs = 1;

	*list = l;
	return 0;
}

static void pakfire_filelist_free(struct pakfire_filelist* list) {
	pakfire_filelist_clear(list);
	pakfire_unref(list->pakfire);
	free(list);
}

PAKFIRE_EXPORT struct pakfire_filelist* pakfire_filelist_ref(struct pakfire_filelist* list) {
	list->nrefs++;

	return list;
}

PAKFIRE_EXPORT struct pakfire_filelist* pakfire_filelist_unref(struct pakfire_filelist* list) {
	if (--list->nrefs > 0)
		return list;

	pakfire_filelist_free(list);
	return NULL;
}

PAKFIRE_EXPORT size_t pakfire_filelist_size(struct pakfire_filelist* list) {
	return list->size;
}

size_t pakfire_filelist_total_size(struct pakfire_filelist* list) {
	struct pakfire_file* file = NULL;
	size_t size = 0;

	for (unsigned int i = 0; i < list->size; i++) {
		file = list->elements[i];

		size += pakfire_file_get_size(file);
	}

	return size;
}

PAKFIRE_EXPORT int pakfire_filelist_is_empty(struct pakfire_filelist* list) {
	return list->size == 0;
}

PAKFIRE_EXPORT void pakfire_filelist_clear(struct pakfire_filelist* list) {
	if (!list->elements)
		return;

	for (unsigned int i = 0; i < list->size; i++)
		pakfire_file_unref(list->elements[i]);

	free(list->elements);
	list->elements = NULL;
	list->elements_size = 0;

	list->size = 0;
}

PAKFIRE_EXPORT struct pakfire_file* pakfire_filelist_get(struct pakfire_filelist* list, size_t index) {
	if (index >= list->size)
		return NULL;

	return pakfire_file_ref(list->elements[index]);
}

PAKFIRE_EXPORT int pakfire_filelist_append(struct pakfire_filelist* list, struct pakfire_file* file) {
	if (!file)
		return EINVAL;

	// Check if we have any space left
	if (list->size >= list->elements_size) {
		int r = pakfire_filelist_grow(list, 64);
		if (r)
			return r;
	}

	list->elements[list->size++] = pakfire_file_ref(file);

	return 0;
}

static int __sort(const void* ptr1, const void* ptr2) {
	struct pakfire_file* file1 = (struct pakfire_file*)ptr1;
	struct pakfire_file* file2 = (struct pakfire_file*)ptr2;

	return pakfire_file_cmp(file1, file2);
}

PAKFIRE_EXPORT void pakfire_filelist_sort(struct pakfire_filelist* list) {
	// Nothing to do on empty list
	if (pakfire_filelist_is_empty(list))
		return;

	qsort(list->elements, list->size, sizeof(*list->elements), __sort);
}

// Returns true if s contains globbing characters
static int is_glob(const char* s) {
	if (strchr(s, '*'))
		return 1;

	if (strchr(s, '?'))
		return 1;

	return 0;
}

static int pakfire_filelist_match_patterns(const char* path, const char** patterns) {
	int flags = 0;
	int r;

	for (const char** pattern = patterns; *pattern; pattern++) {
		// Match any subdirectories
		if (pakfire_string_startswith(path, *pattern))
			return 1;

		// Skip fnmatch if the pattern doesn't have any globbing characters
		if (!is_glob(*pattern))
			continue;

		// Reset flags
		flags = 0;

		/*
			fnmatch is way too eager for patterns line /usr/lib/lib*.so which will also match
			things like /usr/lib/python3.x/blah/libblubb.so.
			To prevent this for absolute file paths, we set the FNM_FILE_NAME flag so that
			asterisk (*) won't match any slashes (/).
		*/
		if (**pattern == '/')
			flags |= FNM_FILE_NAME;

		// Perform matching
		r = fnmatch(*pattern, path, flags);

		// Found a match
		if (r == 0)
			return 1;

		// No match found
		else if (r == FNM_NOMATCH)
			continue;

		// Any other error
		else
			return r;
	}

	// No match
	return 0;
}

int pakfire_filelist_scan(struct pakfire_filelist* list, const char* root,
		const char** includes, const char** excludes) {
	struct pakfire_file* file = NULL;
	struct archive_entry* entry = NULL;
	int r = 1;

	// Root must be absolute
	if (!pakfire_path_is_absolute(root)) {
		errno = EINVAL;
		return 1;
	}

	DEBUG(list->pakfire, "Scanning %s...\n", root);

	if (includes) {
		DEBUG(list->pakfire, "  Includes:\n");

		for (const char** include = includes; *include; include++)
			DEBUG(list->pakfire, "    %s\n", *include);
	}

	if (excludes) {
		DEBUG(list->pakfire, "  Excludes:\n");

		for (const char** exclude = excludes; *exclude; exclude++)
			DEBUG(list->pakfire, "    %s\n", *exclude);
	}

	struct archive* reader = pakfire_make_archive_disk_reader(list->pakfire, 1);
	if (!reader)
		return 1;

	// Allocate a new file entry
	entry = archive_entry_new();
	if (!entry)
		goto ERROR;

	char* paths[] = {
		(char*)root, NULL,
	};

	// Walk through the whole file system tree and find all matching files
	FTS* tree = fts_open(paths, FTS_NOCHDIR, 0);
	if (!tree)
		goto ERROR;

	FTSENT* node = NULL;
	const char* path = NULL;

	while ((node = fts_read(tree))) {
		// Ignore any directories in post order
		if (node->fts_info == FTS_DP)
			continue;

		// Compute the relative path
		path = pakfire_path_relpath(root, node->fts_path);
		if (!path || !*path)
			continue;

		// Skip excludes
		if (excludes && pakfire_filelist_match_patterns(path, excludes)) {
			DEBUG(list->pakfire, "Skipping %s...\n", path);

			r = fts_set(tree, node, FTS_SKIP);
			if (r)
				goto ERROR;

			continue;
		}

		// Skip what is not included
		if (includes && !pakfire_filelist_match_patterns(path, includes)) {
			DEBUG(list->pakfire, "Skipping %s...\n", path);

			// We do not mark the whole tree as to skip because some matches might
			// look for file extensions, etc.
			continue;
		}

		DEBUG(list->pakfire, "Processing %s...\n", path);

		// Reset the file entry
		entry = archive_entry_clear(entry);

		// Set path
		archive_entry_set_pathname(entry, path);

		// Set source path
		archive_entry_copy_sourcepath(entry, node->fts_path);

		// Read all file attributes from disk
		r = archive_read_disk_entry_from_file(reader, entry, -1, node->fts_statp);
		if (r) {
			ERROR(list->pakfire, "Could not read from %s: %m\n", node->fts_path);
			goto ERROR;
		}

		// Create file
		r = pakfire_file_create_from_archive_entry(&file, list->pakfire, entry);
		if (r)
			goto ERROR;

		// Append it to the list
		r = pakfire_filelist_append(list, file);
		if (r) {
			pakfire_file_unref(file);
			goto ERROR;
		}

		pakfire_file_unref(file);
	}

	// Sort the filelist
	pakfire_filelist_sort(list);

	// Success
	r = 0;

ERROR:
	if (entry)
		archive_entry_free(entry);
	archive_read_free(reader);

	return r;
}

int pakfire_filelist_contains(struct pakfire_filelist* list, const char* pattern) {
	if (!pattern) {
		errno = EINVAL;
		return -1;
	}

	for (unsigned int i = 0; i < list->size; i++) {
		struct pakfire_file* file = list->elements[i];

		const char* path = pakfire_file_get_path(file);
		if (!path)
			return -1;

		int r = fnmatch(pattern, path, 0);
		if (r == 0)
			return 1;
	}

	return 0;
}

int pakfire_filelist_walk(struct pakfire_filelist* list,
		const char* pattern, pakfire_filelist_walk_callback callback, void* data) {
	struct pakfire_file* file = NULL;
	int r = 0;

	// Call the callback once for every element on the list
	for (unsigned int i = 0; i < list->size; i++) {
		file = list->elements[i];

		// Skip any files that don't match the pattern
		if (pattern && !pakfire_file_matches(file, pattern))
			continue;

		// Call the callback
		r = callback(list->pakfire, file, data);
		if (r)
			break;
	}

	return r;
}

static int __pakfire_filelist_dump(
		struct pakfire* pakfire, struct pakfire_file* file, void* p) {
	const int debug = *(int*)p;

	char* s = pakfire_file_dump(file);
	if (s) {
		if (debug)
			DEBUG(pakfire, "%s\n", s);
		else
			INFO(pakfire, "%s\n", s);

		free(s);
	}

	return 0;
}

int pakfire_filelist_dump(struct pakfire_filelist* list, int debug) {
#ifndef ENABLE_DEBUG
	// Avoid walking through the entire filelist in non-debug mode
	if (debug)
		return 0;
#endif

	return pakfire_filelist_walk(list, NULL, __pakfire_filelist_dump, &debug);
}

/*
	Verifies all files on the filelist
*/
int pakfire_filelist_verify(struct pakfire_filelist* list, struct pakfire_filelist* errors) {
	struct pakfire_progressbar* progressbar = NULL;
	struct pakfire_file* file = NULL;
	int status;
	int r;

	DEBUG(list->pakfire, "Verifying filelist (%zu file(s))...\n", list->size);

	// Setup progressbar
	r = pakfire_progressbar_create(&progressbar, NULL);
	if (r)
		goto ERROR;

	// Add status
	r = pakfire_progressbar_add_string(progressbar, _("Verifying Files..."));
	if (r)
		goto ERROR;

	// Add bar
	r = pakfire_progressbar_add_bar(progressbar);
	if (r)
		goto ERROR;

	// Add percentage
	r = pakfire_progressbar_add_percentage(progressbar);
	if (r)
		goto ERROR;

	// Add ETA
	r = pakfire_progressbar_add_eta(progressbar);
	if (r)
		goto ERROR;

	// Start the progressbar
	r = pakfire_progressbar_start(progressbar, list->size);
	if (r)
		goto ERROR;

	// Iterate over the entire list
	for (unsigned int i = 0; i < list->size; i++) {
		file = list->elements[i];

		// Verify the file
		r = pakfire_file_verify(file, &status);
		if (r)
			goto ERROR;

		// If the verification failed, we append it to the errors list
		if (status) {
			// Append the file to the error list
			r = pakfire_filelist_append(errors, file);
			if (r)
				goto ERROR;
		}

		// Increment progressbar
		r = pakfire_progressbar_increment(progressbar, 1);
		if (r)
			goto ERROR;
	}

	// Finish
	r = pakfire_progressbar_finish(progressbar);
	if (r)
		goto ERROR;

ERROR:
	if (progressbar)
		pakfire_progressbar_unref(progressbar);

	return r;
}

int pakfire_filelist_cleanup(struct pakfire_filelist* list) {
	int r;

	// Nothing to do if the filelist is empty
	if (pakfire_filelist_is_empty(list))
		return 0;

	// Walk through the list backwards
	for (int i = list->size - 1; i >= 0; i--) {
		r = pakfire_file_cleanup(list->elements[i]);
		if (r)
			return r;
	}

	return 0;
}
