/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2013 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#include <stdlib.h>

#include <solv/pool.h>
#include <solv/solver.h>
#include <solv/util.h>

#include <pakfire/logging.h>
#include <pakfire/package.h>
#include <pakfire/packagelist.h>
#include <pakfire/pakfire.h>
#include <pakfire/private.h>
#include <pakfire/util.h>

#define BLOCK_SIZE 31

struct pakfire_packagelist {
	struct pakfire* pakfire;
	int nrefs;

	struct pakfire_package** elements;
	size_t count;
};

static void pakfire_packagelist_free(struct pakfire_packagelist* list) {
	if (list->elements) {
		for (unsigned int i = 0; i < list->count; i++)
			pakfire_package_unref(list->elements[i]);
		free(list->elements);
	}

	pakfire_unref(list->pakfire);
	free(list);
}

PAKFIRE_EXPORT int pakfire_packagelist_create(
		struct pakfire_packagelist** list, struct pakfire* pakfire) {
	struct pakfire_packagelist* l = calloc(1, sizeof(*l));
	if (!l)
		return 1;

	l->pakfire = pakfire_ref(pakfire);
	l->nrefs = 1;

	*list = l;
	return 0;
}

int pakfire_packagelist_create_from_queue(struct pakfire_packagelist** list, struct pakfire* pakfire, Queue* q) {
	struct pakfire_package* pkg = NULL;

	// Create a new package list
	int r = pakfire_packagelist_create(list, pakfire);
	if (r)
		return r;

	for (int i = 0; i < q->count; i++) {
		r = pakfire_package_create_from_solvable(&pkg, pakfire, q->elements[i]);
		if (r)
			goto ERROR;

		pakfire_packagelist_push(*list, pkg);
		pakfire_package_unref(pkg);
	}

	return 0;

ERROR:
	pakfire_packagelist_free(*list);
	*list = NULL;

	return 1;
}

PAKFIRE_EXPORT struct pakfire_packagelist* pakfire_packagelist_ref(struct pakfire_packagelist* list) {
	list->nrefs++;

	return list;
}

PAKFIRE_EXPORT struct pakfire_packagelist* pakfire_packagelist_unref(struct pakfire_packagelist* list) {
	if (--list->nrefs > 0)
		return list;

	pakfire_packagelist_free(list);
	return NULL;
}

PAKFIRE_EXPORT size_t pakfire_packagelist_count(struct pakfire_packagelist* list) {
	return list->count;
}

static int _packagelist_cmp(const void* pkg1, const void* pkg2) {
	return pakfire_package_cmp(
		*(struct pakfire_package**)pkg1,
		*(struct pakfire_package**)pkg2
	);
}

PAKFIRE_EXPORT void pakfire_packagelist_sort(struct pakfire_packagelist* list) {
	qsort(list->elements, list->count, sizeof(*list->elements), _packagelist_cmp);
}

PAKFIRE_EXPORT struct pakfire_package* pakfire_packagelist_get(struct pakfire_packagelist* list, unsigned int index) {
	if (index < list->count)
		return pakfire_package_ref(list->elements[index]);

	return NULL;
}

PAKFIRE_EXPORT void pakfire_packagelist_push(struct pakfire_packagelist* list, struct pakfire_package* pkg) {
	list->elements = solv_extend(list->elements, list->count, 1, sizeof(pkg), BLOCK_SIZE);
	list->elements[list->count++] = pakfire_package_ref(pkg);
}

int pakfire_packagelist_walk(struct pakfire_packagelist* list,
		pakfire_packagelist_walk_callback callback, void* p) {
	int r = 0;

	for (unsigned int i = 0; i < list->count; i++) {
		r = callback(list->pakfire, list->elements[i], p);
		if (r)
			break;
	}

	return r;
}
